/** 
    To directly compile this project, you'll need G3D 9.00 beta 2 or later
    from http://g3d.sf.net.  The intent of this code is to demonstrate 
    how to use the shaders, however, so you probably can get the value of
    this example without actually building a binary.
*/
#include <G3D/G3DAll.h>

G3D_START_AT_MAIN();

class App : public GApp {
public:

    Shader::Ref     clipShader;

    float clipAngle;
    float clipPos;
    enum Poly {TRIANGLE, QUADRILATERAL} poly;

    App(const GApp::Settings& s) : GApp(s), clipAngle(0), clipPos(0), poly(TRIANGLE) {}

    void loadShader() {
        clipShader = Shader::fromFiles("clip.vrt", "clip.geo", "clip.pix");
    }

    void onInit() {
        loadShader();

        debugWindow->setCaption("Controls");
        GFont::Ref iconFont = GFont::fromFile("icon.fnt");
        GuiLabel* L = debugPane->addLabel("Shape");
        L->setWidth(80);
        debugPane->addRadioButton(GuiText("\x0ea", iconFont, 18), TRIANGLE, &poly, GuiTheme::TOOL_RADIO_BUTTON_STYLE)->moveRightOf(L);
        debugPane->addRadioButton(GuiText("c", iconFont), QUADRILATERAL, &poly, GuiTheme::TOOL_RADIO_BUTTON_STYLE);

        debugPane->addSlider("Angle", &clipAngle, 0.0f, 2.0f * pif());
        debugPane->addSlider("Position", &clipPos, -3.0f, 3.0f);

        debugWindow->setVisible(true);
        showRenderingStats = false;
        developerWindow->setVisible(false);
        developerWindow->cameraControlWindow->setVisible(false);

        defaultCamera.setCoordinateFrame(CFrame::fromXYZYPRDegrees(  2.3f,   0.0f,   5.2f,  25.8f,  -0.6f,   0.0f));

        renderDevice->setColorClearValue(Color3::white());
    }

    bool onEvent(const GEvent& e) {
        if (GApp::onEvent(e)) {
            return true;
        }

        if ((e.type == GEventType::KEY_DOWN) && (e.key.keysym.sym == 'r')) { 
            debugPrintf("Reloading Shader\n");
            loadShader();
            return true; 
        }

        return false;
    }

    void onGraphics3D(RenderDevice* rd, Array<Surface::Ref>& surface) {
        static const Vector3 tri[4]  = {Vector3( 0,  2, 0), Vector3(-2, -2, 0), Vector3( 2, -2, 0), Vector3::zero()};
        static const Vector3 quad[4] = {Vector3(-2,  2, 0), Vector3(-2, -2, 0), Vector3( 2, -2, 0), Vector3(2, 2, 0)};

        const Vector3 clipPlaneNormal(cos(clipAngle), sin(clipAngle), 0.0f);

        ///////////////////////////////////////////////////////////
        // Draw polygon
        const int N = (poly == QUADRILATERAL) ? 4 : 3;
        rd->setObjectToWorldMatrix(Vector3(clipPos, 0, 0));
        const Vector3* v = (N == 4) ? quad : tri; 

        // Enable blending for line smoothing
        rd->setBlendFunc(RenderDevice::BLEND_SRC_ALPHA, RenderDevice::BLEND_ONE_MINUS_SRC_ALPHA);

        rd->pushState(); {
            rd->setCullFace(RenderDevice::CULL_NONE);

            clipShader->args.set("projectionMatrix", rd->invertYMatrix() * rd->projectionMatrix());
            clipShader->args.set("n", clipPlaneNormal, true);
            clipShader->args.set("quadSource", (poly == QUADRILATERAL));
            rd->setShader(clipShader);

            clipShader->args.set("color", Color3(0.6f, 0.8f, 1.0f));
            for (int i = 0; i < 2; ++i) {
                rd->beginPrimitive(PrimitiveType::TRIANGLES); {
                    rd->setTexCoord(0, v[3]); // lastVertex
                    for (int j = 0; j < 3; ++j) {
                        rd->sendVertex(v[j]);
                    }
                } rd->endPrimitive();

                // Render wireframe
                clipShader->args.set("color", Color3::black());
                rd->setLineWidth(3.0f);
                rd->setRenderMode(RenderDevice::RENDER_WIREFRAME);
                rd->setPolygonOffset(-0.5f);
            }
        } rd->popState();

        ///////////////////////////////////////////////////////////
        // Show vertices and edges

        for (int j = 0; j < N; ++j) {
            Draw::sphere(Sphere(v[j], 0.1f), rd, Color3::brown(), Color4::clear());
        }

        rd->setLineWidth(4.0f);
        rd->setColor(Color3::brown());
        glEnable(GL_LINE_STIPPLE);
        glLineStipple(1, 0x00FF);
        rd->beginPrimitive(PrimitiveType::LINE_STRIP);
        for (int j = 0; j <= N; ++j) {
            rd->sendVertex(v[j % N]);
        }
        rd->endPrimitive();
        glDisable(GL_LINE_STIPPLE);

        //////////////////////////////////////////////////////////
        // Draw plane
        rd->setObjectToWorldMatrix(CFrame());
        Draw::plane(Plane(clipPlaneNormal, Vector3::zero()), rd, Color4::clear(), Color3::gray());
        Draw::arrow(Vector3::zero(), clipPlaneNormal, rd, Color3::gray());

    }
};

int main(const int argc, const char* argv[]) {

#   ifdef G3D_WIN32
        if (FileSystem::exists("data-files")) {
            chdir("data-files");
        }
#   endif

    GApp::Settings s;
    s.window.refreshRate = 60;
    s.window.asynchronous = false;
    s.film.enabled = false;
    s.window.msaaSamples = 4;
    s.window.caption = "Clipper - Morgan McGuire, NVIDIA / Williams College";

    App(s).run();

    return 0;
}
